#include <math.h>
#include <stdio.h>
#include <stdlib.h>
#include <ctype.h>
#include <float.h>
#include <string.h>
#include <limits>
#include <time.h>

#include "lr1_problem.h"
#include "externfuncs.h"
#include "misc.h"

extern double dgnrm2(int n, double *x, double *xl, double *xu, double *g);
extern void spcg(int n, double *x, double *xl, double *xu, lr1_problem* prob, double *g, double *D, double delta, double rtol, double *s,  int *cgiters, int *info, double *Hs);
extern void cauchy(int n, double *x, double *xl, double *xu, lr1_problem* prob, double *g, double *D, double delta, double *alpha, double *s, double *Hs);

void tron(lr1_problem *prob, double *w, int *iter, char* debugfile)
{
	/* Parameters for updating the iterates. */
	double eta0 = 1e-4, eta1 = 0.25, eta2 = 0.75;
	/* Parameters for updating the trust region size delta. */
	double sigma1 = 0.25, sigma2 = 0.5, sigma3 = 4;

	int l = prob->l;
	int n = prob->n;
	int N = 2*n;
	int i; 
	double delta, snorm; //one=1.0, p5=0.5;
	double alpha, alphac = 1, f, fc, prered, actred, gs;
	int search = 1, inc = 1, max_iter;	
	double gnorm, gnorm0;
	int status, cgiters;

	double *wl = Malloc(double, N);
	double *wu = Malloc(double, N);
	double *s = Malloc(double, N);
	double *wc = Malloc(double, N);
	double *Hs = Malloc(double, N);

	double *g = Malloc(double, N);
	double *z = Malloc(double, l); 
	double *D = Malloc(double, l);
	double *C = Malloc(double, l);
	
	/* initial the output for info() */
	info_initial(1, debugfile);

	/* initial constraints wl and wu  */
	for(i=0; i<N; ++i)
	{
		wl[i] = 0;
		wu[i] = std::numeric_limits<double>::infinity(); // or max()?
	}

	for (i=0; i<N; i++)
		w[i] = 0;

	f = prob->func(w, z);
	prob->grad(z, g, D);
	
	gnorm0 = dnrm2_(&N, g, &inc);
	delta = 1000*gnorm0; 
	gnorm = dgnrm2(N, w, wl, wu, g);

	if (gnorm < prob->gtol * gnorm0) 
		search = 0;

	max_iter = 10000;
	*iter = 0;

	double timer = 0.0;
	double start = clock();

	while( (*iter)<=max_iter && search )
	{
		//info("iteration: %2d\n", *iter+1);
		fc = f;
		memcpy(wc, w, sizeof(double)*N);

		//D and g come from the line before the while(first iter) or last iteration.
		//solve sub-problem by conjugate gradient method
		cauchy(N, w, wl, wu, prob, g, D, delta, &alphac, s, Hs); 
		spcg(N, w, wl, wu, prob, g, D, delta, prob->cgtol, s, &cgiters, &status, Hs);

		f = prob->func(w, z);
		/*
		if(f > prob->maxfev){
			info("f:%lf > maxfev %lf\n", f, prob->maxfev);
			break;
		}*/

		gs = ddot_(&N, g, &inc, s, &inc);//NOTICE: old g
		prered = -(0.5*ddot_(&N, s, &inc, Hs, &inc) + gs);

		/* Compute the actual reduction. */
		actred = fc - f; 
                                    
		/* On the first iteration, adjust the initial step bound. */
		snorm = dnrm2_(&N, s, &inc);
		if ( (*iter) == 0)
			delta = min(delta, snorm);
			   
		/* Compute prediction alpha*snorm of the step. */
		if (f - fc - gs <= 0)
			alpha = sigma3;
		else
			alpha = max(sigma1, -0.5*(gs/(f - fc - gs)));

		/* Update the trust region bound according to the ratio
		of actual to predicted reduction. */
		if (actred < eta0*prered){ // Reduce delta. Step is not successful.
			delta = min(max(alpha, sigma1)*snorm, sigma2*delta);
		}
		else{
			/* Reduce delta. Step is not sufficiently successful. */
			if (actred < eta1*prered) 
				delta = max(sigma1*delta, min(alpha*snorm, sigma2*delta));
			else{ 
				/* The ratio of actual to predicted reduction is in
				   the interval (eta1,eta2). We are allowed to either
				   increase or decrease delta. */
				if (actred < eta2*prered)
					delta = max(sigma1*delta, min(alpha*snorm, sigma3*delta));
				/* The ratio of actual to predicted reduction exceeds eta2.
				   Do not decrease delta. */
				else
					delta = max(delta, min(alpha*snorm, sigma3*delta));
			}
		}

		info("iter %2d terminate\nact %5.3e pre %5.3e delta %5.3e f %5.8e |g| %5.8e CG %3d Hvtime %lf nHv %d\n", *iter+1, actred, prered, delta, f, gnorm, cgiters, (double)prob->totalHvtime/CLOCKS_PER_SEC, prob->numberofHv);

/*
#if DEBUG & 2
		info("w:\n");
		for(i=0; i<2*n; ++i)
			info("%lf ", w[i]);
		info("\n");
#endif
*/

		if ( actred > eta0*prered )
		{
			(*iter)++;
			prob->grad(z, g, D); //Notice that z is from the last fun
			gnorm = dgnrm2(N, w, wl, wu, g); // g is computed when "prered" is derived 

			//XXX start of testing
			timer += clock()-start;

			double gc_norm = prob->grad_condiction(g, w);
			int nnz = 0;
			for(i=0; i<n; ++i)
			{
				double wi = w[i]-w[i+n];
				if(wi != 0.0)
					nnz ++;
			}
			printf("iter %d time %lf f %lf accuracy %lf Gnorm %lf\t", *iter, timer/CLOCKS_PER_SEC, f, prob->testing(w),gc_norm);
			printf("sparse %lf nonzero %d n %d\n", (double)(n-nnz)/n, nnz, n);

			start = clock();
			//XXX end of testing

			if ( gnorm <= prob->gtol*gnorm0 )
			{
				info("gnorm :%lf <= gtol*%lf\n", gnorm, gnorm0);
				break;
			}
		}
		else
		{
			memcpy(w, wc, sizeof(double)*N);
			f = fc;
		}

		if(f < prob->fmin)
		{
			info("warning: f < %lf\n", prob->fmin);
			break;
		}
		if(fabs(actred)<=prob->fatol && fabs(prered)<=prob->fatol)
		{
			info("convergence: fatol test satisfied\n");
			break;
		}
		if(fabs(actred)<=prob->frtol*fabs(f) && fabs(prered)<=prob->frtol*fabs(f))
		{
			info("convergence: frtol test satisfied\n");
			break;
		}
		info_flush();
	}

	if ((*iter) >= max_iter)
		info("Exceed max_iter %d\n", max_iter);
	info_terminate();

	free(wl);
	free(wu);
	free(s);
	free(wc);
	free(Hs);
	free(g);
	free(z);
	free(D);
	free(C);
}

